#include "mainform.h"
#include <QFocusEvent>
#include <QTextCodec>
#include <QLibraryInfo>

#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>

#include <fcntl.h>
#if defined(_WIN32)
#include <windows.h>

#define O_LARGEFILE     0
#endif

#include "any.h"

void BlockSizeBox :: focusOutEvent ( QFocusEvent * event )
{
	long bs, bs0;
	bs = bs0 = currentText().toLong();

	int log = 0;
	while (bs>1) { log++; bs>>=1; }

	long v = 1<<log;
	if ( (bs0-v) > v/2 ) v<<=1;

	if (v<512) v=512;
	if (v>65536) v=65536;

	QString vs; vs.setNum(v);
	setEditText(vs);

	QComboBox::focusOutEvent(event);
}

void DiskButton::mousePressEvent ( QMouseEvent * event )
{
	QPushButton::mousePressEvent(event);

	MainWindow* mainw = ( (MainWindow *) (parent()->parent()) );
	QFileDialog fd(this);
	QString ImageName = fd.getOpenFileName();
	if ( !ImageName.isNull() )
		mainw->diskBox->setEditText( ImageName );

	QFocusEvent Qf(QEvent::FocusOut);
	QCoreApplication::sendEvent(mainw->diskBox, &Qf);
}

void MyThread::run()
{
	MainWindow* mainw = ( (MainWindow *) (parent()->parent()) );
	setTerminationEnabled();
	mainw->StartSurrect();
}

void StartButton::mousePressEvent ( QMouseEvent * event )
{
	QPushButton::mousePressEvent(event);

	MainWindow* mainw = ( (MainWindow *) (parent()->parent()) );

	if (mainw->state == 0)
	{
		mainw->state = 1;
		mainw->checkState();

		mainw->thread = new MyThread((QWidget*) parent());
		mainw->thread->start();
	}
	else
	{
		mainw->newAction();
	}
}

void SaveSelectedButton::mousePressEvent ( QMouseEvent * event )
{
	QPushButton::mousePressEvent(event);

	MainWindow* mainw = ( (MainWindow *) (parent()->parent()) );
	mainw->SaveSelected();
}

void MainWindow::currentItemChanged ( QTreeWidgetItem * current,
		QTreeWidgetItem * previous )
{
	if (!previous)
	{
		if (image_access)
			saveSelectedButton->setEnabled(true);
	}

	if (!current)
	{
		saveSelectedButton->setEnabled(false);
	}
}

void MainWindow::editTextChanged (const QString & text)
{
	checkState();

	char *filename = diskBox->currentText().toAscii().data();
	filename[diskBox->currentText().length()]='\0';
#if defined(__linux__)
	if ( access(filename, R_OK) )
	{
		startButton->setEnabled(false);
		saveSelectedButton->setEnabled(false);
		image_access = false;
		return;
	}

	struct stat st;
	stat(filename, &st);
	if ( !S_ISREG(st.st_mode) && !S_ISBLK(st.st_mode) )
	{
		startButton->setEnabled(false);
		saveSelectedButton->setEnabled(false);
		image_access = false;
		return;
	}

	image_access = true;
	startButton->setEnabled(true);
	if ( treeWidget->currentItem() )
	{
		saveSelectedButton->setEnabled(true);
	}
#elif defined _WIN32
	int fd = open(filename, O_RDONLY | O_LARGEFILE);
	if ( fd == -1 )
	{
		startButton->setEnabled(false);
		saveSelectedButton->setEnabled(false);
		image_access = false;
		return;
	}

	::close(fd);

	image_access = true;
	startButton->setEnabled(true);
	if ( treeWidget->currentItem() )
		saveSelectedButton->setEnabled(true);
#endif
}

int MainWindow::checkState()
{
	switch(state)
	{
		case 0:
			startButton->setText(_("Start"));
			startButton->setStatusTip(_("Press to start device scanning"));
			startButton->setToolTip(_("Start device scanning"));
			break;

		case 1:
		case 2:
			startButton->setText(_("New"));
			startButton->setStatusTip(_("Press to begin new recovering loop..."));
			startButton->setToolTip(_("Create new inode table"));
			break;
	}

	return 0;
}

MainWindow::MainWindow()
{
	setObjectName(QString::fromUtf8("MainWindow"));
	setLayoutDirection(Qt::LeftToRight);
	actionOpen = new QAction(this);
	actionOpen->setObjectName(QString::fromUtf8("actionOpen"));
	actionSave = new QAction(this);
	actionSave->setObjectName(QString::fromUtf8("actionSave"));
	actionSave_As = new QAction(this);
	actionSave_As->setObjectName(QString::fromUtf8("actionSave_As"));
	actionExit = new QAction(this);
	actionExit->setObjectName(QString::fromUtf8("actionExit"));
	actionNew = new QAction(this);
	actionNew->setObjectName(QString::fromUtf8("actionNew"));
	actionAbout = new QAction(this);
	actionAbout->setObjectName(QString::fromUtf8("actionAbout"));
	actionAbout_Qt = new QAction(this);
	actionAbout_Qt->setObjectName(QString::fromUtf8("actionAbout_Qt"));
	actionHelp = new QAction(this);
	actionHelp->setObjectName(QString::fromUtf8("actionHelp"));
	centralwidget = new QWidget(this);
	centralwidget->setObjectName(QString::fromUtf8("centralwidget"));
	centralwidget->setLayoutDirection(Qt::LeftToRight);
	gridLayout = new QGridLayout(centralwidget);
	gridLayout->setSpacing(6);
	gridLayout->setMargin(9);
	gridLayout->setObjectName(QString::fromUtf8("gridLayout"));
	saveSelectedButton = new SaveSelectedButton(centralwidget);
	saveSelectedButton->setObjectName(QString::fromUtf8("saveSelectedButton"));
	saveSelectedButton->setGeometry(QRect(415, 484, 129, 23));

	gridLayout->addWidget(saveSelectedButton, 7, 3, 1, 1);

	startButton = new StartButton(centralwidget);
	startButton->setObjectName(QString::fromUtf8("startButton"));
	startButton->setGeometry(QRect(280, 484, 129, 23));

	gridLayout->addWidget(startButton, 7, 2, 1, 1);

	diskButton = new DiskButton(centralwidget);
	diskButton->setObjectName(QString::fromUtf8("diskButton"));
	diskButton->setGeometry(QRect(415, 35, 129, 23));
	diskButton->setLayoutDirection(Qt::LeftToRight);

	gridLayout->addWidget(diskButton, 1, 3, 1, 1);

	progressBar = new QProgressBar(centralwidget);
	progressBar->setObjectName(QString::fromUtf8("progressBar"));
	progressBar->setGeometry(QRect(9, 457, 535, 21));
	progressBar->setOrientation(Qt::Horizontal);
	progressBar->setInvertedAppearance(false);

	gridLayout->addWidget(progressBar, 6, 0, 1, 4);

	listoffileslabel = new QLabel(centralwidget);
	listoffileslabel->setObjectName(QString::fromUtf8("listoffileslabel"));
	listoffileslabel->setGeometry(QRect(9, 90, 129, 20));

	gridLayout->addWidget(listoffileslabel, 3, 0, 1, 1);

	blocksizeLabel = new QLabel(centralwidget);
	blocksizeLabel->setObjectName(QString::fromUtf8("blocksizeLabel"));
	blocksizeLabel->setGeometry(QRect(9, 9, 129, 20));

	gridLayout->addWidget(blocksizeLabel, 0, 0, 1, 1);

	firstblockBox = new QCheckBox(centralwidget);
	firstblockBox->setObjectName(QString::fromUtf8("firstblockBox"));
	firstblockBox->setGeometry(QRect(9, 65, 129, 18));

	gridLayout->addWidget(firstblockBox, 2, 0, 1, 1);

	progressLabel = new QLabel(centralwidget);
	progressLabel->setObjectName(QString::fromUtf8("progressLabel"));
	progressLabel->setGeometry(QRect(9, 431, 129, 20));

	gridLayout->addWidget(progressLabel, 5, 0, 1, 1);

	blocksizeBox = new BlockSizeBox(centralwidget);
	blocksizeBox->setObjectName(QString::fromUtf8("blocksizeBox"));
	blocksizeBox->setGeometry(QRect(9, 36, 129, 21));
	blocksizeBox->setEditable(true);

	gridLayout->addWidget(blocksizeBox, 1, 0, 1, 1);

	deviceLabel = new QLabel(centralwidget);
	deviceLabel->setObjectName(QString::fromUtf8("deviceLabel"));
	deviceLabel->setGeometry(QRect(144, 9, 130, 20));

	gridLayout->addWidget(deviceLabel, 0, 1, 1, 1);

	diskBox = new DiskBox(centralwidget);
	diskBox->setObjectName(QString::fromUtf8("diskBox"));
	diskBox->setGeometry(QRect(144, 36, 265, 21));
	diskBox->setEditable(true);

	gridLayout->addWidget(diskBox, 1, 1, 1, 2);

	treeWidget = new QTreeWidget(centralwidget);
	treeWidget->setObjectName(QString::fromUtf8("treeWidget"));
	treeWidget->setGeometry(QRect(10, 110, 531, 311));

	gridLayout->addWidget(treeWidget, 4, 0, 1, 4);

	setCentralWidget(centralwidget);
	menubar = new QMenuBar(this);
	menubar->setObjectName(QString::fromUtf8("menubar"));
	menubar->setGeometry(QRect(0, 0, 553, 25));
	menu_2 = new QMenu(menubar);
	menu_2->setObjectName(QString::fromUtf8("menu_2"));
	menu = new QMenu(menubar);
	menu->setObjectName(QString::fromUtf8("menu"));
	setMenuBar(menubar);
	statusbar = new QStatusBar(this);
	statusbar->setObjectName(QString::fromUtf8("statusbar"));
	setStatusBar(statusbar);
	blocksizeLabel->setBuddy(blocksizeBox);
	deviceLabel->setBuddy(diskBox);
	QWidget::setTabOrder(blocksizeBox, diskBox);
	QWidget::setTabOrder(diskBox, diskButton);
	QWidget::setTabOrder(diskButton, firstblockBox);
	QWidget::setTabOrder(firstblockBox, startButton);
	QWidget::setTabOrder(startButton, saveSelectedButton);

	menubar->addAction(menu->menuAction());
	menubar->addAction(menu_2->menuAction());
	menu_2->addAction(actionHelp);
	menu_2->addSeparator();
	menu_2->addAction(actionAbout);
	menu_2->addAction(actionAbout_Qt);
	menu->addAction(actionNew);
	menu->addAction(actionOpen);
	menu->addAction(actionSave);
	menu->addAction(actionSave_As);
	menu->addSeparator();
	menu->addAction(actionExit);

	retranslateUi();
	init();

	QSize size(553, 559);
	size = size.expandedTo(this->minimumSizeHint());
	resize(size);


	QMetaObject::connectSlotsByName(this);
} // setupUi

void MainWindow::closeEvent(QCloseEvent * event)
{
	if ( NewSurrect() )
	{
		event->ignore();
		return;
	}

	DeInitSurrect();
	event->accept();
}

bool MainWindow::maybeSave()
{
	QMessageBox::StandardButton ret;
	ret = QMessageBox::warning(this, _("Application"),
			_("The document has been modified.\n"
				"Do you want to save your changes?"),
			QMessageBox::Save | QMessageBox::Discard | QMessageBox::Cancel);
	if (ret == QMessageBox::Save)
		return saveAction();
	else if (ret == QMessageBox::Cancel)
		return false;

	return true;
}

void MainWindow::retranslateUi()
{
#if defined(__linux__)
	QTextCodec::setCodecForCStrings(
			QTextCodec::codecForName ("KOI8-R") );
#elif defined(_WIN32)
	QTextCodec::setCodecForCStrings(
			QTextCodec::codecForName ("CP1251") );
#endif

	setWindowTitle(_("anysurreqt"));
	
	actionOpen->setText(_("&Open..."));
	actionOpen->setStatusTip(_("Open inode table..."));
	
	actionSave->setText(_("&Save"));
	actionSave->setStatusTip(_("Save inode table"));
	
	actionSave_As->setText(_("Save &As..."));
	actionSave_As->setStatusTip(_("Save inode table with new name..."));
	
	actionExit->setText(_("E&xit..."));
	actionExit->setStatusTip(_("Close anysurreqt..."));
	
	actionNew->setText(_("&New"));
	actionNew->setStatusTip(_("New inode table..."));
	
	actionAbout->setText(_("About..."));
	actionAbout->setStatusTip(_("About anysurreqt..."));
	
	actionAbout_Qt->setText(_("About Qt..."));
	actionAbout_Qt->setStatusTip(_("About Qt library..."));
	
	actionHelp->setText(_("Help..."));
	actionHelp->setStatusTip(_("Help..."));
	
	diskBox->setStatusTip(_("Enter existing filename or device name"));
	diskBox->setToolTip(_("Device or image filename"));
	
	firstblockBox->setText(_("Force from first block"));
	firstblockBox->setStatusTip(_("Check it for scanning device from begin"));
	firstblockBox->setToolTip(_("Force from first block checkbox"));
	
	saveSelectedButton->setText(_("Save Selected..."));
	saveSelectedButton->setStatusTip(_("Select file or directory in tree and press this button..."));
	saveSelectedButton->setToolTip(_("Save files selected in tree"));
	
	diskButton->setText(_("Image..."));
	diskButton->setStatusTip(_("Use for selecting image file with open file dialog"));
	diskButton->setToolTip(_("Select Image file..."));
	
	deviceLabel->setText(_("Device or image file:"));
	
	progressLabel->setText(_("Progress:"));
	
	blocksizeLabel->setText(_("BlockSize:"));
	
	listoffileslabel->setText(_("List of Files"));
	
	progressBar->setFormat(_("%p% (%v/%m)"));
	progressBar->setToolTip(_("Device scanning progress..."));
	
	menu_2->setTitle(_("&Help"));
	menu->setTitle(_("&File"));

	blocksizeBox->setStatusTip(_("Enter blocksize of device filesystem"));
	blocksizeBox->setToolTip(_("Filesystem blocksize"));

	treeWidget->setStatusTip(_("Select files for recovering..."));
} // retranslateUi

void MainWindow::about()
{
	QMessageBox::about(this, _("About anysurreqt"),
			_("anysurreqt -- is graphical version of anysurrect,\n" \
				"program for recovering files from devices\n" \
				"with damaged filesystem."));
}

void MainWindow::help()
{
	char wd[1024];
	getcwd(wd, 1024);
	assistantClient->showPage(QString(wd) + QString("/docs/index.html"));
}

void MainWindow::init()
{
	treeWidget->setColumnCount(11);
	QStringList headstr;
	headstr << _("Name") << _("Size") << _("Type") << _("Link/Device") << 
		_("Mode") << _("User ID") << _("Group ID") <<
		_("Access Time") << _("Modification Time") << 
		_("Status Change Time") << _("Links") <<
		_("Inode Number");
	QTreeWidgetItem *head = new QTreeWidgetItem(headstr);
	treeWidget->setHeaderItem(head);

	QIntValidator *val = new QIntValidator(512, 65536, centralwidget);
	blocksizeBox->setValidator (val);

	blocksizeBox->addItem("4096");
	blocksizeBox->addItem("2048");
	blocksizeBox->addItem("1024");
	blocksizeBox->addItem("512");

#if defined(__linux__)
	diskBox->addItem("/dev/hda15");
#else
	char buf[1024];
	buf[0]=buf[1]='\0';
	GetLogicalDriveStrings(1024, buf);
	char *p = buf;
	while (p[0])
	{
		QString str(p);
		str.chop(1);
		diskBox->addItem( "\\\\.\\"/*"*/ + str );
		p += strlen(p)+1;
	}
#endif

	editTextChanged(diskBox->currentText());

	QFocusEvent Qf(QEvent::FocusOut);

	saveSelectedButton->setEnabled(false);

	QObject::connect(treeWidget, SIGNAL(currentItemChanged(QTreeWidgetItem *,
					QTreeWidgetItem *)),
			this, SLOT(currentItemChanged(QTreeWidgetItem *,
					QTreeWidgetItem *)));

	QObject::connect(diskBox, SIGNAL(editTextChanged (const QString &)),
			this, SLOT(editTextChanged (const QString &)));

	qRegisterMetaType<uint32_t>("uint32_t");
	QObject::connect(this, SIGNAL(SigAddItem(any_sb_info *,            
					const char *, uint32_t, uint32_t)),
			this, SLOT(AddItem (any_sb_info *,            
					const char *, uint32_t, uint32_t)), 
			Qt::QueuedConnection);
	QObject::connect(this, SIGNAL(SigMessage(QString, QString)),
			this, SLOT(Message(QString, QString)), 
			Qt::QueuedConnection);
	QObject::connect(this, SIGNAL(SetProgress (int)),
			progressBar, SLOT(setValue (int)));
	QObject::connect(this, SIGNAL(SigFinishSurrect ()),
			this, SLOT(FinishSurrect ()));
	QObject::connect(actionSave, SIGNAL(triggered()),
		       	this, SLOT(saveAction()));
	QObject::connect(actionSave_As, SIGNAL(triggered()),
		       	this, SLOT(saveAsAction()));
	QObject::connect(actionOpen, SIGNAL(triggered()),
		       	this, SLOT(openAction()));
	QObject::connect(actionNew, SIGNAL(triggered()),
		       	this, SLOT(newAction()));

	QObject::connect(actionHelp, SIGNAL(triggered()), this, SLOT(help()));
	QObject::connect(actionAbout, SIGNAL(triggered()), this, SLOT(about()));
	QObject::connect(actionAbout_Qt, SIGNAL(triggered()), qApp, SLOT(aboutQt()));
	QObject::connect(actionExit, SIGNAL(triggered()), qApp, SLOT(closeAllWindows()));

	assistantClient = new QAssistantClient(QLibraryInfo::location(QLibraryInfo::BinariesPath), this);

	QStringList arguments;
	arguments << "-profile" << QString("docs") + QDir::separator() + QString("anysurreqt.adp");
	assistantClient->setArguments(arguments);

	state = 0;
	saved = 0;
	FD = -1;
	sb_info = NULL;
	block_bitmap = NULL;
	thread = NULL;
	wait_terminate = 0;
	save_state = 0;
	open_state = 0;

	actionNew->setEnabled(0);
	actionSave->setEnabled(0);
	actionSave_As->setEnabled(0);

	checkState();

	InitSurrect();
}
